# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for the 'zhmc nic' command group.
"""

import urllib3
import pytest
# pylint: disable=unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=unused-import

from .utils import run_zhmc, parse_output, pick_test_resources, fixup

urllib3.disable_warnings()


# Expected output properties for 'nic list' when --names-only option is specif.
NIC_LIST_PROPNAMES_NAMES_ONLY = [
    "cpc",
    "partition",
    "name",
]

# Expected output properties for 'nic list' when no other options are specified
NIC_LIST_PROPNAMES_DEFAULT = NIC_LIST_PROPNAMES_NAMES_ONLY + [
    "description",
    "type",
]

# Expected output properties for 'nic list' when --uri option is specified
NIC_LIST_PROPNAMES_URI = NIC_LIST_PROPNAMES_DEFAULT + [
    "element-uri",
]

# Expected output properties for 'nic list' when --all option is specified.
# Note: These are the z15 properties for a non-SSC-mgmt NIC that are common
# for all backing adapter types. Additional properties may be shown.
NIC_LIST_PROPNAMES_ALL = [
    "class",
    "description",
    "device-number",
    "element-id",
    "element-uri",
    "mac-address",
    "name",
    "parent",
    "ssc-management-nic",
    "type",
    "vlan-id",
]

# Expected output properties for 'nic show'.
# Note: These are the z15 properties for a non-SSC-mgmt NIC that are common
# for all backing adapter types. Additional properties may be shown.
NIC_SHOW_PROPNAMES = [
    "class",
    "description",
    "device-number",
    "element-id",
    "element-uri",
    "mac-address",
    "network-adapter-name",
    "network-adapter-port-index",
    "network-adapter-port-name",
    "name",
    "parent",
    "parent-name",
    "ssc-management-nic",
    "type",
    "vlan-id",
]

# Properties that should be skipped when verifying values
# (e.g. because they are volatile or have string conversion issues)
NIC_PROPNAMES_SKIP = [
    "device-number",  # CSV does not retain leading zeros
]

# Properties that need to be forced to string type in CSV tests
NIC_PROPNAMES_CSV_STR = [
    "description",  # May be empty string which by default becomes None
    "device-number",  # CSV does not retain leading zeros
]


def add_backing_adapter_properties(nic, properties):
    """
    Add artificial properties for the backing network adapter of the nic to the
    properties.
    """
    try:
        vswitch_uri = nic.get_property('virtual-switch-uri')
    except KeyError:
        pass
    else:
        vswitch_props = nic.manager.session.get(vswitch_uri)
        adapter_uri = vswitch_props['backing-adapter-uri']
        cpc = nic.manager.parent.manager.parent
        adapter = cpc.adapters.resource_object(adapter_uri)
        properties['network-adapter-name'] = adapter.name
        port_index = vswitch_props['port']
        properties['network-adapter-port-index'] = port_index
        port = adapter.ports.find(index=port_index)
        properties['network-adapter-port-name'] = port.name

    try:
        port_uri = nic.get_property('network-adapter-port-uri')
    except KeyError:
        pass
    else:
        port_props = nic.manager.session.get(port_uri)
        properties['network-adapter-port-name'] = port_props['name']
        properties['network-adapter-port-index'] = port_props['index']
        adapter_props = nic.manager.session.get(port_props['parent'])
        properties['network-adapter-name'] = adapter_props['name']


TESTCASES_NIC_LIST = [
    # Testcases for test_nic_list()
    # Each list item is a testcase, which is a tuple with items:
    # - options (list of str): Options for 'zhmc nic list' command
    # - exp_props (list of str): Expected property names in output
    # - exact (bool): If True, the output property names must be exactly the
    #   expected properties. If False, additional properties are allowed.
    (
        ['--names-only'],
        NIC_LIST_PROPNAMES_NAMES_ONLY,
        True,
    ),
    (
        [],
        NIC_LIST_PROPNAMES_DEFAULT,
        True,
    ),
    (
        ['--uri'],
        NIC_LIST_PROPNAMES_URI,
        True,
    ),
    (
        ['--all'],
        NIC_LIST_PROPNAMES_ALL,
        False,
    ),
]


@pytest.mark.parametrize(
    "options, exp_props, exact",
    TESTCASES_NIC_LIST
)
@pytest.mark.parametrize(
    "out_format",
    ['json', 'csv']
)
def test_nic_list(
        out_format, options, exp_props, exact, dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test 'nic list' command.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        for partition in pick_test_resources(cpc.partitions.list()):

            exp_nics = partition.nics.list()
            exp_nics_props = {nic.name:
                              dict(nic.properties) for nic in exp_nics}
            # Add artificial properties
            for nic in exp_nics:
                nic_props = exp_nics_props[nic.name]
                nic_props['cpc'] = cpc.name
                nic_props['partition'] = partition.name

            args = ['-o', out_format, 'nic', 'list', cpc.name, partition.name] \
                + options
            rc, stdout, stderr = run_zhmc(args)

            assert rc == 0, (
                f"Unexpected rc={rc} (expected 0),\n"
                f"stdout:\n{stdout}\n"
                f"stderr:\n{stderr}\n"
            )
            assert stderr == ""

            nic_list = parse_output(stdout, out_format, list)

            assert len(nic_list) == len(exp_nics_props)
            for nic_item in nic_list:
                assert 'name' in nic_item
                nic_name = nic_item['name']
                assert nic_name in exp_nics_props
                exp_nic_props = exp_nics_props[nic_name]
                for pname in exp_props:
                    assert pname in nic_item
                    if pname not in NIC_PROPNAMES_SKIP:
                        exp_value = exp_nic_props.get(pname, 'undefined')
                        if out_format == 'csv':
                            force_type = \
                                str if pname in NIC_PROPNAMES_CSV_STR else None
                            value = fixup(nic_item[pname], force_type)
                        else:
                            value = nic_item[pname]
                        assert value == exp_value, (
                            f"Unexpected value for property {pname} of "
                            f"NIC {nic_name} in partition {partition.name} on "
                            f"CPC {cpc.name}:\n"
                            f"Expected: {exp_value!r}\n"
                            f"Actual: {value!r}"
                        )
                if exact:
                    assert len(nic_item) == len(exp_props)


@pytest.mark.parametrize(
    "nic_type",
    ["osd", "iqd", "cna", "roce", "osh", "neth", "netd"]
)
@pytest.mark.parametrize(
    "out_format",
    ['json', 'csv']
)
def test_nic_show(
        out_format, nic_type, dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test 'nic show' command.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        # In the first call to this test function, list the NICs of all
        # partitions of the CPC and store the result on the CPC object,
        # for fast retrieval in the subsequent calls to this test function.
        if hasattr(cpc, 'nic_list'):
            nic_list = getattr(cpc, 'nic_list')
        else:
            nic_list = []
            for partition in cpc.partitions.list():
                for nic in partition.nics.list():
                    nic_list.append(nic)
            setattr(cpc, 'nic_list', nic_list)

        # Select the NICS of the desired type
        typed_nic_list = []
        for nic in nic_list:
            if nic.prop('type') == nic_type:
                typed_nic_list.append(nic)
        if not typed_nic_list:
            pytest.skip(f"The partitions on CPC {cpc.name} do not have any "
                        f"NICs of type {nic_type}")

        for nic in pick_test_resources(typed_nic_list):
            partition = nic.manager.parent
            cpc = partition.manager.parent

            rc, stdout, stderr = run_zhmc(
                ['-o', out_format, 'nic', 'show', cpc.name, partition.name,
                 nic.name])

            assert rc == 0, (
                f"Unexpected rc={rc} (expected 0),\n"
                f"stdout:\n{stdout}\n"
                f"stderr:\n{stderr}\n"
            )
            assert stderr == ""

            nic_props = parse_output(stdout, out_format, dict)

            exp_nic_props = dict(nic.properties)
            exp_nic_props['parent-name'] = partition.name
            add_backing_adapter_properties(nic, exp_nic_props)

            for pname in NIC_SHOW_PROPNAMES:
                assert pname in nic_props
                if pname not in NIC_PROPNAMES_SKIP:
                    exp_value = exp_nic_props.get(pname, 'undefined')
                    if out_format == 'csv':
                        force_type = \
                            str if pname in NIC_PROPNAMES_CSV_STR else None
                        value = fixup(nic_props[pname], force_type)
                    else:
                        value = nic_props[pname]
                    assert value == exp_value, (
                        f"Unexpected value for property {pname} of "
                        f"NIC {nic.name} in partition {partition.name} on "
                        f"CPC {cpc.name}:\n"
                        f"Expected: {exp_value!r}\n"
                        f"Actual: {value!r}"
                    )


# TODO: Add test function for 'nic create'
# TODO: Add test function for 'nic update'
# TODO: Add test function for 'nic delete'



# Copyright 2016,2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for NICs.
"""

import time

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props
from ._cmd_partition import find_partition
from ._cmd_cpc import find_cpc


# Defaults for NIC creation
SSC_IP_ADDRESS_TYPES = ['ipv4', 'ipv6', 'linklocal', 'dhcp']
VLAN_TYPES = ['enforced', 'none']


def find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name):
    """
    Find a NIC by name and return its resource object.
    """
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)
    try:
        nic = partition.nics.find(name=nic_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return nic


@cli.group('nic', options_metavar=COMMAND_OPTIONS_METAVAR)
def nic_group():
    """
    Command group for managing NICs (DPM mode only).

    The commands in this group work only on CPCs that are in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@nic_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.option('--type', is_flag=True, required=False, hidden=True)
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def nic_list(cmd_ctx, cpc, partition, **options):
    """
    List the NICs in a partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_list(cmd_ctx, cpc, partition, options))


@nic_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('NIC', type=str, metavar='NIC')
@click.pass_obj
def nic_show(cmd_ctx, cpc, partition, nic):
    """
    Show the details of a NIC.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Partition.
      - 'network-adapter-name' - Name of the backing Adapter
      - 'network-adapter-port-name' - Name of the backing Adapter Port
      - 'network-adapter-port-index' - Index of the backing Adapter Port

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_show(cmd_ctx, cpc, partition, nic))


@nic_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.option('--name', type=str, required=True,
              help='The name of the new NIC.')
@click.option('--description', type=str, required=False,
              help='The description of the new NIC. '
              'Default: empty')
@click.option('--adapter', type=str, required=False,
              help='The name of the network adapter with the port backing the '
              'new NIC. Required.')
@click.option('--port', type=str, metavar='NAME|INDEX', required=False,
              help='The name or index of the network port backing the new NIC. '
              'Required.')
@click.option('--device-number', type=str, required=False,
              help='The device number to be used for the new NIC. '
              'Default: auto-generated')
@click.option('--ssc-management-nic', type=bool, required=False,
              help='Indicates that this NIC should be used as a management '
              'NIC for Secure Service Container to access the web interface. '
              'Only applicable to NICs of SSC partitions. '
              'Default: False')
@click.option('--ssc-ip-address-type', type=click.Choice(SSC_IP_ADDRESS_TYPES),
              required=False,
              help='Secure Service Container IP address type. '
              'Only applicable to and required for NICs of '
              'SSC partitions.')
@click.option('--ssc-ip-address', type=str, required=False,
              help='IP Address of the SSC management web interface. '
              'Only applicable to and required for NICs of SSC '
              'partitions when ssc-ip-address-type is ipv4 or ipv6.')
@click.option('--ssc-mask-prefix', type=str, required=False,
              help='Network Mask of the SSC management NIC. '
              'Only applicable to and required for NICs of SSC '
              'partitions when ssc-ip-address-type is ipv4 or ipv6.')
@click.option('--vlan-id', metavar='[INT|none]', type=str, required=False,
              help='VLAN ID of the NIC, or "none" for setting no VLAN ID. '
              'On z14 or later CPCs, specifying a VLAN ID requires using the '
              '--vlan-type option. '
              'Only applicable to management NICs of SSC partitions, and to '
              'OSA and Hipersocket NICs of non-SSC partitions. '
              'Default: No VLAN ID')
@click.option('--vlan-type', required=False, type=click.Choice(VLAN_TYPES),
              help='VLAN type of the NIC, or "none" for setting no VLAN type. '
              'Only supported on z14 or later CPCs, and required if specifying '
              'a VLAN ID with the --vlan-id option. '
              'Default: No VLAN type')
@click.pass_obj
def nic_create(cmd_ctx, cpc, partition, **options):
    """
    Create a NIC in a partition.

    For all types of network adapters and all CPC/SE generations, the backing
    adapter and port can be specified with the --adapter and --port options.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_create(cmd_ctx, cpc, partition,
                                               options))


@nic_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('NIC', type=str, metavar='NIC')
@click.option('--name', type=str, required=False,
              help='The new name of the NIC.')
@click.option('--description', type=str, required=False,
              help='The new description of the NIC.')
@click.option('--adapter', type=str, required=False,
              help='The name of the network adapter with the port backing the '
              'NIC.')
@click.option('--port', type=str, metavar='NAME|INDEX', required=False,
              help='The name or index of the network port backing the NIC.')
@click.option('--device-number', type=str, required=False,
              help='The new device number to be used for the NIC.')
@click.option('--ssc-management-nic', type=bool, required=False,
              help='Indicates that this NIC should be used as a management '
              'NIC for Secure Service Container to access the web interface. '
              'Only applicable to NICs of SSC partitions. ')
@click.option('--ssc-ip-address-type', type=click.Choice(SSC_IP_ADDRESS_TYPES),
              required=False,
              help='Secure Service Container IP address type. '
              'Only applicable to NICs of SSC partitions.')
@click.option('--ssc-ip-address', type=str, required=False,
              help='IP Address of the SSC management web interface. '
              'Only applicable to NICs of SSC partitions.')
@click.option('--ssc-mask-prefix', type=str, required=False,
              help='Network Mask of the SSC management NIC. '
              'Only applicable to NICs of SSC partitions.')
@click.option('--vlan-id', metavar='[INT|none]', type=str, required=False,
              help='VLAN ID of the NIC, or "none" for setting no VLAN ID. '
              'On z14 or later CPCs, specifying a VLAN ID requires using the '
              '--vlan-type option. '
              'Only applicable to management NICs of SSC partitions, and to '
              'OSA and Hipersocket NICs of non-SSC partitions.')
@click.option('--vlan-type', required=False, type=click.Choice(VLAN_TYPES),
              help='VLAN type of the NIC, or "none" for setting no VLAN type. '
              'Only supported on z14 or later CPCs, and required if specifying '
              'a VLAN ID with the --vlan-id option.')
@click.pass_obj
def nic_update(cmd_ctx, cpc, partition, nic, **options):
    """
    Update the properties of a NIC.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    If the backing adapter and port needs to be updated, it can be specified
    with the --adapter and --port options, for all types of network adapters
    and all CPC/SE generations.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_update(cmd_ctx, cpc, partition, nic,
                                               options))


@nic_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('NIC', type=str, metavar='NIC')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the NIC.',
              prompt='Are you sure you want to delete this NIC ?')
@click.pass_obj
def nic_delete(cmd_ctx, cpc, partition, nic):
    """
    Delete a NIC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_nic_delete(cmd_ctx, cpc, partition, nic))


def backing_uri(cmd_ctx, cpc, org_options, required=False):
    """
    Determine the backing adapter port or vswitch to be used from the
    --adapter and --port options, and return a dict with the correct property
    for the URI of the backing port or vswitch to be used for the "Create NIC"
    or "Update NIC Properties" operation.

    If none of the --adapter and --port options are specified, None is
    returned.

    If only one of these options is specified, or if they are required but
    not specified, an exception is raised.

    Returns:
      dict with the backing URI property set, if --adapter and --port options
      are specified. None, if these options are not required and not specified.

    Raises:
      click exception for various error situations.
    """

    if bool(org_options['adapter']) != bool(org_options['port']):
        raise click_exception(
            "The --adapter and --port options must be specified both or none.",
            cmd_ctx.error_format)

    if not org_options['adapter'] and not org_options['port']:
        if required:
            raise click_exception(
                "The --adapter and --port options are required but were not "
                "specified.", cmd_ctx.error_format)
        return None

    adapter_name = org_options['adapter']
    port_name = org_options['port']
    try:
        adapter = cpc.adapters.find(name=adapter_name)
    except zhmcclient.NotFound:
        raise click_exception(
            "Could not find adapter '{a}' in CPC '{c}'.".
            format(a=adapter_name, c=cpc.name),
            cmd_ctx.error_format)
    try:
        # Try interpreting the --port value as a port name
        port = adapter.ports.find(name=port_name)
    except zhmcclient.NotFound:
        # Try interpreting the --port value as a port index
        try:
            port_index = int(port_name)
        except ValueError:
            raise click_exception(
                "Could not find port with name '{p}' on "
                "adapter '{a}' in CPC '{c}'.".
                format(p=port_name, a=adapter_name, c=cpc.name),
                cmd_ctx.error_format)
        try:
            port = adapter.ports.find(index=port_index)
        except zhmcclient.NotFound:
            raise click_exception(
                "Could not find port with name or index '{p}' on "
                "adapter '{a}' in CPC '{c}'.".
                format(p=port_name, a=adapter_name, c=cpc.name),
                cmd_ctx.error_format)

    try:
        nes_feature = cpc.api_feature_enabled('network-express-support')
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    adapter_family = adapter.get_property('adapter-family')
    if adapter_family in ('osa', 'hipersockets') and not nes_feature:
        # a vswitch-based NIC (OSA, HS up to z16)
        port_index = port.get_property('index')
        filter_args = {
            'backing-adapter-uri': adapter.uri,
            'port': port_index,
        }
        try:
            vswitch = cpc.virtual_switches.find(**filter_args)
        except zhmcclient.NotFound:
            raise click_exception(
                "Could not find virtual switch with backing adapter '{a}' "
                "and port index '{p}' in CPC '{c}'.".
                format(a=adapter.name, p=port_index, c=cpc.name),
                cmd_ctx.error_format)
        return {'virtual-switch-uri': vswitch.uri}

    # an adapter-based NIC (RoCE, CNA up to z16 or all adapter types
    # since z17)
    return {'network-adapter-port-uri': port.uri}


def cmd_nic_list(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    filter_args = build_filter_args(cmd_ctx, options['filter'])
    try:
        nics = partition.nics.list(filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['type']:
        click.echo("The --type option is deprecated and type information "
                   "is now always shown.")

    show_list = [
        'name',
        'cpc',
        'partition',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    cpc_additions = {}
    partition_additions = {}
    for nic in nics:
        cpc_additions[nic.uri] = cpc_name
        partition_additions[nic.uri] = partition_name
    additions = {
        'cpc': cpc_additions,
        'partition': partition_additions,
    }

    sort_props = build_sort_props(cmd_ctx, options['sort'], default=['name'])
    try:
        print_resources(cmd_ctx, nics, cmd_ctx.output_format, show_list,
                        additions, all=options['all'], sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_nic_show(cmd_ctx, cpc_name, partition_name, nic_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    nic = find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name)

    try:
        nic.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(nic.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = partition_name

    # Add artificial properties in case of a vswitch-based NIC (OSA, HS
    # up to z16)
    try:
        vswitch_uri = nic.get_property('virtual-switch-uri')
    except KeyError:
        pass
    else:
        vswitch_props = client.session.get(vswitch_uri)
        adapter_uri = vswitch_props['backing-adapter-uri']
        cpc = find_cpc(cmd_ctx, client, cpc_name)
        adapter = cpc.adapters.resource_object(adapter_uri)
        properties['network-adapter-name'] = adapter.name

        port_index = vswitch_props['port']
        properties['network-adapter-port-index'] = port_index

        port = adapter.ports.find(index=port_index)
        properties['network-adapter-port-name'] = port.name

    # Add artificial properties in case of an adapter-based NIC (RoCE, CNA
    # up to z16 or all adapter types since z17)
    try:
        port_uri = nic.get_property('network-adapter-port-uri')
    except KeyError:
        pass
    else:
        port_props = client.session.get(port_uri)
        properties['network-adapter-port-name'] = port_props['name']
        properties['network-adapter-port-index'] = port_props['index']

        adapter_props = client.session.get(port_props['parent'])
        properties['network-adapter-name'] = adapter_props['name']

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def nic_mitigate_name(cmd_ctx, nic, desired_name):
    """
    Mitigate the naming issue by explicitly naming the NIC, if needed.
    """
    cmd_ctx.spinner.stop()
    click.echo("Mitigating HMC defect: Updating name of NIC '{n}'.".
               format(n=desired_name))

    attempts = 3  # number of attempts, including initial one
    sleep_time = 0.5  # seconds, between retries
    while True:
        nic.update_properties({'name': desired_name})
        time.sleep(sleep_time)
        nic.pull_properties('name')
        if nic.name == desired_name:
            break
        # The name is not as desired yet, retry
        attempts -= 1
        if attempts > 0:
            click.echo("Mitigating HMC defect: Retrying after {s} s, because "
                       "NIC name is still different.".
                       format(s=sleep_time))
            continue
        raise click_exception(
            "HMC defect mitigation for updating the NIC name did not work. "
            "NIC name is: {n}".format(n=nic.name),
            cmd_ctx.error_format)


def nic_mitigate_uuid(cmd_ctx, prior_nic_uris, partition, desired_name):
    """
    Mitigate the UUID issue by finding the created NIC.
    """
    cmd_ctx.spinner.stop()
    click.echo("Mitigating HMC defect: Ignoring HTTP 500,0 "
               "'Invalid UUID string:' when creating NIC '{n}'.".
               format(n=desired_name))

    attempts = 3  # number of attempts, including initial one
    sleep_time = 0.5  # seconds, between retries
    while True:
        partition.pull_properties("nic-uris")
        new_nics = [nic for nic in partition.nics.list()
                    if nic.uri not in prior_nic_uris]
        if len(new_nics) == 1:
            break
        if len(new_nics) > 1:
            new_nic_names = [nic.name for nic in new_nics]
            raise click_exception(
                "HMC defect mitigation for 500,0 did not work, because there "
                "is more than one new NIC: {n}".format(n=new_nic_names),
                cmd_ctx.error_format)
        # We got no new NICs yet, retry
        attempts -= 1
        if attempts > 0:
            click.echo("Mitigating HMC defect: Retrying after {s} s, because "
                       "no new NICs were returned.".format(s=sleep_time))
            time.sleep(sleep_time)
            continue
        raise click_exception(
            "HMC defect mitigation for 500,0 did not work, because no new "
            "NICs were returned after retrying.", cmd_ctx.error_format)

    new_nic = new_nics[0]
    return new_nic


def cmd_nic_create(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    # Flags to control whether HMC defects are mitigated
    cpc = partition.manager.parent
    nes_feature = cpc.api_feature_enabled('network-express-support')
    mitigate_uuid = nes_feature  # HTTP 500,0 "Invalid UUID string:"
    mitigate_name = nes_feature  # NIC name is not set

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
        'port': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    set_vlan_id_type(cmd_ctx, properties, org_options)

    properties.update(backing_uri(
        cmd_ctx, partition.manager.cpc, org_options, required=True))

    if mitigate_uuid:
        prior_nics = partition.nics.list()
        prior_nic_uris = [nic.uri for nic in prior_nics]

    try:
        new_nic = partition.nics.create(properties)
    except zhmcclient.HTTPError as exc:
        if mitigate_uuid and exc.http_status == 500 and exc.reason == 0 and \
                "Invalid UUID string:" in exc.stack:
            # pylint: disable=possibly-used-before-assignment
            new_nic = nic_mitigate_uuid(
                cmd_ctx, prior_nic_uris, partition, options['name'])
        else:
            raise click_exception(exc, cmd_ctx.error_format)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if mitigate_name:
        new_nic.pull_properties('name')
        if new_nic.name != options['name']:
            nic_mitigate_name(cmd_ctx, new_nic, options['name'])

    cmd_ctx.spinner.stop()
    click.echo("New NIC '{n}' has been created.".
               format(n=new_nic.properties['name']))


def cmd_nic_update(cmd_ctx, cpc_name, partition_name, nic_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    nic = find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name)

    # Flags to control whether HMC defects are mitigated
    partition = nic.manager.parent
    cpc = partition.manager.parent
    nes_feature = cpc.api_feature_enabled('network-express-support')
    mitigate_name = nes_feature  # NIC name is not set

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
        'port': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    set_vlan_id_type(cmd_ctx, properties, org_options)

    uri_prop = backing_uri(
        cmd_ctx, nic.manager.partition.manager.cpc, org_options)
    if uri_prop:
        properties.update(uri_prop)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating NIC '{n}'.".
                   format(n=nic_name))
        return

    try:
        nic.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if mitigate_name and options['name']:
        nic.pull_properties('name')
        if nic.name != options['name']:
            nic_mitigate_name(cmd_ctx, nic, options['name'])

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != nic_name:
        click.echo("NIC '{n}' has been renamed to '{nn}' and was updated.".
                   format(n=nic_name, nn=properties['name']))
    else:
        click.echo(f"NIC '{nic_name}' has been updated.")


def cmd_nic_delete(cmd_ctx, cpc_name, partition_name, nic_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    nic = find_nic(cmd_ctx, client, cpc_name, partition_name, nic_name)

    try:
        nic.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"NIC '{nic_name}' has been deleted.")


def set_vlan_id_type(cmd_ctx, properties, org_options):
    """
    Set the 'vlan-id' and 'vlan-type' properties from the options.
    """
    vlan_id = org_options['vlan-id']
    if vlan_id is None:
        pass  # No change on update, use HMC default on create
    elif vlan_id == 'none':
        # Reset to no VLAN ID (important for update)
        properties['vlan-id'] = None
    else:
        try:
            properties['vlan-id'] = int(vlan_id)
        except ValueError:
            raise click_exception(
                "Invalid value for '--vlan-id': {} is not a valid integer".
                format(vlan_id), cmd_ctx.error_format)

    vlan_type = org_options['vlan-type']
    if vlan_type is None:
        pass  # No change on update, use HMC default on create
    elif vlan_type == 'none':
        # Reset to no VLAN type (important for update)
        properties['vlan-type'] = None
    else:
        properties['vlan-type'] = vlan_type



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for partition links in DPM mode.
"""


import click
from click_option_group import optgroup

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, FILTER_OPTIONS, \
    build_filter_args, SORT_OPTIONS, build_sort_props, ALL_PARTITION_STATUSES
from ._cmd_cpc import find_cpc, find_cpc_by_uri
from ._cmd_adapter import find_adapter_by_uri


# Defaults for partition link creation
PARTITIONLINK_TYPES = ['hipersockets', 'smc-d']  # 'ctc' is not yet supported
DEFAULT_PARTITIONLINK_TYPE = 'hipersockets'
DEFAULT_SMCD_STARTING_FID = 4096
DEFAULT_HS_STARTING_DEVNO = 7400
DEFAULT_HS_MTU_SIZE = 8
# DEFAULT_CTC_DEVICES_PER_PATH = 4


def find_partitionlink(cmd_ctx, client, partitionlink_name):
    """
    Find a partition link by name and return its resource object.
    """
    console = client.consoles.console
    try:
        partitionlink = console.partition_links.find(name=partitionlink_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    return partitionlink


@cli.group('partitionlink', options_metavar=COMMAND_OPTIONS_METAVAR)
def partitionlink_group():
    """
    Command group for managing partition links (DPM mode only).

    The commands in this group work only on CPCs that are in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@partitionlink_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='[CPC]', required=False)
@add_options(LIST_OPTIONS)
@add_options(FILTER_OPTIONS)
@add_options(SORT_OPTIONS)
@click.pass_obj
def partitionlink_list(cmd_ctx, cpc, **options):
    """
    List the partition links associated with a CPC, or all partition links.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_partitionlink_list(cmd_ctx, cpc, options))


@partitionlink_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PARTITIONLINK', type=str, metavar='PARTITIONLINK')
@click.pass_obj
def partitionlink_show(cmd_ctx, partitionlink):
    # pylint: disable=line-too-long
    """
    Show the details of a partition link.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'adapter-name' - Name of the backing adapter, for SMC-D and Hipersocket.
      - 'bus-connections.nics.nic-name' - Names of the NICs, for SMC-D and Hipersocket.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """  # noqa: E501
    cmd_ctx.execute_cmd(
        lambda: cmd_partitionlink_show(cmd_ctx, partitionlink))


@partitionlink_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@optgroup.group('General options')
@optgroup.option('--name', type=str, required=True,
                 help='The name of the new partition link.')
@optgroup.option('--type', type=click.Choice(PARTITIONLINK_TYPES),
                 required=False, default=DEFAULT_PARTITIONLINK_TYPE,
                 help='The type of the new partition link. Default: {d}'.
                 format(d=DEFAULT_PARTITIONLINK_TYPE))
@optgroup.option('--description', type=str, required=False,
                 help='The description of the new partition link.')
@optgroup.group('For type=hipersockets')
@optgroup.option('--starting-devno', type=str, required=False,
                 help='The starting number for the device number assignment, '
                 'acting as a default. If no device number for a NIC is '
                 'provided when adding a partition, the next available '
                 'device number starting from this value will be used. '
                 'Default: {d}'.format(d=DEFAULT_HS_STARTING_DEVNO))
@optgroup.option('--mtu-size', type=int, required=False,
                 help='The maximum transmission unit (MTU) size of the '
                 'Hipersockets adapter that will be created. The maximum frame '
                 'size is implied by this value. Valid values are: '
                 '8 - 8 KB MTU size and 16 KB maximum frame size; '
                 '16 - 16 KB MTU size and 24 KB maximum frame size; '
                 '32 - 32 KB MTU size and 40 KB maximum frame size; '
                 '56 - 56 KB MTU size and 64 KB maximum frame size. '
                 'Default: {d}'.format(d=DEFAULT_HS_MTU_SIZE))
@optgroup.group('For type=smc-d')
@optgroup.option('--starting-fid', type=int, required=False,
                 help='The starting number for the Functional ID (FID) '
                 'assignment, acting as a default. If no FID for a NIC is '
                 'provided when adding a partition, the next available FID '
                 'starting from this value will be used. '
                 'Default: {d}'.format(d=DEFAULT_SMCD_STARTING_FID))
# @optgroup.group('For type=ctc')
# @optgroup.option('--devices-per-path', type=int, required=False,
#                  help='The number of devices to be created per path when '
#                  'adding a partition. '
#                  'Default: {d}'.format(d=DEFAULT_CTC_DEVICES_PER_PATH))
@click.pass_obj
def partitionlink_create(cmd_ctx, cpc, **options):
    """
    Create a partition link associated with a CPC.

    This command cannot be used to have partitions attached to the new
    partition link.

    This command supports only the creation of partitions links of type SMC-D
    and Hipersockets; type CTC is not supported.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_partitionlink_create(cmd_ctx, cpc, options))


@partitionlink_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PARTITIONLINK', type=str, metavar='PARTITIONLINK')
@optgroup.group('General options')
@optgroup.option('--name', type=str, required=False,
                 help='The new name of the partition link.')
@optgroup.option('--description', type=str, required=False,
                 help='The new description of the partition link.')
@optgroup.group('For type=hipersockets')
@optgroup.option('--starting-devno', type=str, required=False,
                 help='The new starting number for the device number '
                 'assignment, acting as a default. If no device number for a '
                 'NIC is provided when adding a partition, the next available '
                 'device number starting from this value will be used.')
@optgroup.option('--mtu-size', type=int, required=False,
                 help='The new maximum transmission unit (MTU) size for the '
                 'Hipersockets adapter. The maximum frame '
                 'size is implied by this value. Valid values are: '
                 '8 - 8 KB MTU size and 16 KB maximum frame size; '
                 '16 - 16 KB MTU size and 24 KB maximum frame size; '
                 '32 - 32 KB MTU size and 40 KB maximum frame size; '
                 '56 - 56 KB MTU size and 64 KB maximum frame size.')
@optgroup.group('For type=smc-d')
@optgroup.option('--starting-fid', type=int, required=False,
                 help='The new starting number for the Functional ID (FID) '
                 'assignment, acting as a default. If no FID for a NIC is '
                 'provided when adding a partition, the next available FID '
                 'starting from this value will be used.')
# @optgroup.group('For type=ctc')
# @optgroup.option('--devices-per-path', type=int, required=False,
#                  help='Number of devices to be created per path when adding '
#                  'a partition. '
#                  'Default: {d}'.format(d=DEFAULT_CTC_DEVICES_PER_PATH))
@click.pass_obj
def partitionlink_update(cmd_ctx, partitionlink, **options):
    """
    Update the properties of a partition link.

    This command cannot be used to attach or detach partitions to or from
    a partition link.

    The partition link type and the associated CPC cannot be changed once a
    partition link has been created.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_partitionlink_update(cmd_ctx, partitionlink, options))


@partitionlink_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PARTITIONLINK', type=str, metavar='PARTITIONLINK')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the partition link.',
              prompt='Are you sure you want to delete this partition link ?')
@click.pass_obj
def partitionlink_delete(cmd_ctx, partitionlink):
    """
    Delete a partition link.

    For Hipersocket partition links, this will cause the NICs that belong to
    the partition link to be deleted in the linked partitions.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_partitionlink_delete(cmd_ctx, partitionlink))


@partitionlink_group.command(
    'list-partitions', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('PARTITIONLINK', type=str, metavar='PARTITIONLINK')
@add_options(LIST_OPTIONS)
@click.option('--name', type=str, required=False,
              help="Regular expression filter to limit the returned partitions "
              "to those with a matching name.")
@click.option('--status', type=str, required=False,
              help="Filter to limit the returned partitions to those with a "
              "matching status. Valid status values are: {sv}.".
              format(sv=', '.join(ALL_PARTITION_STATUSES)))
@add_options(SORT_OPTIONS)
@click.pass_obj
def partitionlink_list_partitions(cmd_ctx, partitionlink, **options):
    """
    List the partitions attached to a partition link.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_partitionlink_list_partitions(
            cmd_ctx, partitionlink, options))


@partitionlink_group.command(
    'attach-partition', options_metavar=COMMAND_OPTIONS_METAVAR,
    hidden=True)  # TODO: Implement in zhmcclient
@click.argument('PARTITIONLINK', type=str, metavar='PARTITIONLINK')
@click.option('--partition', type=str, required=True,
              help='The name of the partition.')
@click.pass_obj
def partitionlink_attach_partition(cmd_ctx, partitionlink, **options):
    """
    Attach a partition to a partition link.

    For Hipersocket partition links, this will cause one or more NICs to be
    created in the attached partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_partitionlink_attach_partition(
            cmd_ctx, partitionlink, options))


@partitionlink_group.command(
    'detach-partition', options_metavar=COMMAND_OPTIONS_METAVAR,
    hidden=True)  # TODO: Implement in zhmcclient
@click.argument('PARTITIONLINK', type=str, metavar='PARTITIONLINK')
@click.option('--partition', type=str, required=True,
              help='The name of the partition.')
@click.pass_obj
def partitionlink_detach_partition(cmd_ctx, partitionlink, **options):
    """
    Detach a partition from a partition link.

    For Hipersocket partition links, this will cause the NICs that belong to
    the partition link to be deleted in the detached partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_partitionlink_detach_partition(
            cmd_ctx, partitionlink, options))


def cmd_partitionlink_list(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'cpc',
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'state',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])
    # TODO: Add adapter-name in case of --all
    # if options['all']:
    #     show_list.extend([
    #         'adapter-name',
    #     ])

    # Make sure to include only properties in additional properties
    # that are known resource properties and are not included in the
    # default set of properties for a list operation.
    standard_props = ['cpc', 'name', 'type', 'state', 'object-uri']
    additional_props = [p for p in show_list if p not in standard_props]

    filter_args = build_filter_args(cmd_ctx, options['filter'])

    if cpc_name:  # None if not specified
        cpc = find_cpc(cmd_ctx, client, cpc_name)
        if filter_args is None:
            filter_args = {}
        filter_args['cpc-uri'] = cpc.uri

    try:
        partitionlinks = console.partition_links.list(
            additional_properties=additional_props, filter_args=filter_args)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    # Prepare the additions dict of dicts. It contains additional
    # (=non-resource) property values by property name and by resource URI.
    additions = {}
    updates = {}

    # Add artificial property 'cpc':
    additions['cpc'] = {}
    for pl in partitionlinks:
        cpc_uri = pl.get_property('cpc-uri')
        cpc = find_cpc_by_uri(cmd_ctx, client, cpc_uri)
        additions['cpc'][pl.uri] = cpc.name

    if options['all']:
        additions['adapter-name'] = {}
        updates['bus-connections'] = {}
        for pl in partitionlinks:

            # Add artificial property 'adapter-name':
            adapter_uri = pl.prop('adapter-uri', None)
            if adapter_uri:  # Exists only for SMC-D and Hipersocket type
                cpc_uri = pl.get_property('cpc-uri')
                cpc = find_cpc_by_uri(cmd_ctx, client, cpc_uri)
                adapter = find_adapter_by_uri(cmd_ctx, cpc, adapter_uri)
                additions['adapter-name'][pl.uri] = adapter.name

            # Add artificial property 'nic-name' in 'bus-connections.nics':
            bc_list = pl.prop('bus-connections', None)
            if bc_list:  # Exists only for SMC-D and Hipersocket type
                updates_bc_list = []
                for bc_item in bc_list:
                    nic_items = bc_item['nics']
                    updates_nics = []
                    for nic_item in nic_items:
                        nic_uri = nic_item['nic-uri']
                        nic_props = client.session.get(nic_uri)
                        updates_nics.append({'nic-name': nic_props['name']})
                    updates_bc_list.append({'nics': updates_nics})
                updates['bus-connections'][pl.uri] = updates_bc_list

    sort_props = build_sort_props(cmd_ctx, options['sort'],
                                  default=['cpc', 'name'])

    try:
        print_resources(
            cmd_ctx, partitionlinks, cmd_ctx.output_format, show_list,
            additions, all=options['all'], sort_props=sort_props,
            updates=updates)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_partitionlink_show(cmd_ctx, partitionlink_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partitionlink = find_partitionlink(cmd_ctx, client, partitionlink_name)

    try:
        partitionlink.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(partitionlink.properties)

    # Add artificial property 'adapter-name':
    adapter_uri = properties.get('adapter-uri', None)
    if adapter_uri:  # Exists only for SMC-D and Hipersocket type
        cpc_uri = properties['cpc-uri']
        cpc = find_cpc_by_uri(cmd_ctx, client, cpc_uri)
        adapter = find_adapter_by_uri(cmd_ctx, cpc, adapter_uri)
        properties['adapter-name'] = adapter.name

    # Add artificial property 'nic-name' in 'bus-connections.nics':
    bc_list = properties.get('bus-connections', None)
    if bc_list:  # Exists only for SMC-D and Hipersocket type
        for bc_item in bc_list:
            nic_items = bc_item['nics']
            for nic_item in nic_items:
                nic_uri = nic_item['nic-uri']
                nic_props = client.session.get(nic_uri)
                nic_item['nic-name'] = nic_props['name']

    # # Hide some long or deeply nested properties in table output formats.
    # if not options['all'] and cmd_ctx.output_format in TABLE_FORMATS:
    #     hide_property(properties, 'xxxxxxxxx')

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_partitionlink_create(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name_map = {
        'starting-devno': 'starting-device-number',
        'mtu-size': 'maximum-transmission-unit-size',
    }

    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    properties['cpc-uri'] = cpc.uri

    try:
        new_partitionlink = console.partition_links.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New partition link '{p}' has been created.".
               format(p=new_partitionlink.properties['name']))


def cmd_partitionlink_update(cmd_ctx, partitionlink_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partitionlink = find_partitionlink(cmd_ctx, client, partitionlink_name)

    name_map = {
        'starting-devno': 'starting-device-number',
        'mtu-size': 'maximum-transmission-unit-size',
    }

    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating partition link '{p}'.".
                   format(p=partitionlink_name))
        return

    try:
        partitionlink.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != partitionlink_name:
        click.echo("Partition link '{pl}' has been renamed to '{pln}' and was "
                   "updated.".
                   format(pl=partitionlink_name, pln=properties['name']))
    else:
        click.echo(f"Partition link '{partitionlink_name}' has been updated.")


def cmd_partitionlink_delete(cmd_ctx, partitionlink_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partitionlink = find_partitionlink(cmd_ctx, client, partitionlink_name)

    try:
        partitionlink.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo(f"Partition link '{partitionlink_name}' has been deleted.")


def cmd_partitionlink_list_partitions(cmd_ctx, partitionlink_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partitionlink = find_partitionlink(cmd_ctx, client, partitionlink_name)

    show_list = [
        'cpc',
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'status',
            'description',
            'nics',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    name_filter = options['name']
    status_filter = options['status']
    try:
        partitions = partitionlink.list_attached_partitions(
            name=name_filter, status=status_filter)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    # Prepare the additions dict of dicts. It contains additional
    # (=non-resource) property values by property name and by resource URI.
    additions = {}

    # Add artificial property 'cpc':
    additions['cpc'] = {}
    for p in partitions:
        cpc = p.manager.parent
        additions['cpc'][p.uri] = cpc.name

    if partitionlink.get_property('type') in ('smc-d', 'hipersockets'):
        # Add artificial property 'nics':
        additions['nics'] = {}
        for p in partitions:
            bc_list = partitionlink.get_property('bus-connections')
            for bc_item in bc_list:
                if bc_item['partition-name'] == p.name:
                    nic_items = bc_item['nics']
                    additions['nics'][p.uri] = []
                    for nic_item in nic_items:
                        nic_uri = nic_item['nic-uri']
                        nic_props = client.session.get(nic_uri)
                        additions['nics'][p.uri].append(nic_props['name'])

    sort_props = build_sort_props(cmd_ctx, options['sort'],
                                  default=['cpc', 'name'])

    try:
        print_resources(
            cmd_ctx, partitions, cmd_ctx.output_format, show_list,
            additions, all=options['all'], sort_props=sort_props)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_partitionlink_attach_partition(cmd_ctx, partitionlink_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partitionlink = find_partitionlink(cmd_ctx, client, partitionlink_name)

    cpc_uri = partitionlink.get_property('cpc-uri')
    cpc = client.cpcs.find(**{'object-uri': cpc_uri})

    partition_name = options['partition']  # Required
    try:
        partition = cpc.partitions.find(name=partition_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    try:
        partitionlink.attach_partition(partition)  # TODO: Implement
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Partition '{p}' was attached to partition link '{pl}'.".
               format(p=partition_name, pl=partitionlink.name))


def cmd_partitionlink_detach_partition(cmd_ctx, partitionlink_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partitionlink = find_partitionlink(cmd_ctx, client, partitionlink_name)

    cpc_uri = partitionlink.get_property('cpc-uri')
    cpc = client.cpcs.find(**{'object-uri': cpc_uri})

    partition_name = options['partition']  # Required
    try:
        partition = cpc.partitions.find(name=partition_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    try:
        partitionlink.detach_partition(partition)  # TODO: Implement
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Partition '{p}' was detached from partition link '{pl}'.".
               format(p=partition_name, pl=partitionlink.name))



# Copyright 2017,2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Function tests for global options, using faked HMCs (zhmcclient mock).
"""

import os
import subprocess  # nosec: B404
import re
import json
import pytest
from zhmcclient import Client
from zhmcclient_mock import FakedSession

from zhmccli._helper import convert_ec_mcl_description

from .utils import call_zhmc_child, call_zhmc_inline, assert_rc, \
    assert_patterns

TEST_LOGFILE = 'tmp_testfile.log'


def setup_faked_hmc():
    """
    Set up a faked HMC with two CPC resources.

    Returns:
      FakedSession: The faked session.
    """
    faked_session = FakedSession(
        'fake-host', 'hmc1', '2.16', '4.10',
        userid='fake-user')

    faked_session.hmc.consoles.add({
        'object-id': None,
        # object-uri will be automatically set
        'parent': None,
        'class': 'console',
        'name': 'fake-console1',
        'description': 'Console #1',
        "ec-mcl-description": {
            "ec": [
                {
                    "number": "P41499",
                    "part-number": "01LL062",
                    "mcl": [
                        {
                            "level": "333",
                            "type": "retrieved",
                            "last-update": 1650524126000
                        },
                        {
                            "level": "333",
                            "type": "activated",
                            "last-update": 1650526481000
                        },
                        {
                            "level": "327",
                            "type": "accepted",
                            "last-update": 1642750400000
                        },
                        {
                            "level": "333",
                            "type": "installable-concurrent",
                            "last-update": None
                        },
                        {
                            "level": "328",
                            "type": "removable-concurrent",
                            "last-update": None
                        }
                    ],
                    "description": "Hardware Management Console Framework",
                    "type": "SYSTEM"
                },
                {
                    "number": "P41470",
                    "part-number": "01LL058",
                    "mcl": [
                        {
                            "level": "2",
                            "type": "retrieved",
                            "last-update": 1566551669000
                        },
                        {
                            "level": "2",
                            "type": "activated",
                            "last-update": 1566551779000
                        },
                        {
                            "level": "2",
                            "type": "accepted",
                            "last-update": 1566553286000
                        },
                        {
                            "level": "2",
                            "type": "installable-concurrent",
                            "last-update": None
                        },
                        {
                            "level": "000",
                            "type": "removable-concurrent",
                            "last-update": None
                        }
                    ],
                    "description": "Hardware Management Console Platform Fw",
                    "type": "HMCBIOS"
                }
            ]
        }
    })

    faked_session.hmc.cpcs.add({
        'object-id': 'fake-cpc1-oid',
        # object-uri is set up automatically
        'parent': None,
        'class': 'cpc',
        'name': 'CPC1',
        'description': 'CPC 1 (classic mode)',
        'status': 'operating',
        'dpm-enabled': False,
        'se-version': '2.16',
        'machine-type': '3931',
        'machine-model': 'A01',
        'machine-serial-number': 'SN-CPC1',
    })
    faked_session.hmc.cpcs.add({
        'object-id': 'fake-cpc2-oid',
        # object-uri is set up automatically
        'parent': None,
        'class': 'cpc',
        'name': 'CPC2',
        'description': 'CPC 2 (DPM mode)',
        'status': 'active',
        'dpm-enabled': True,
        'se-version': '2.16',
        'machine-type': '3931',
        'machine-model': 'LA1',
        'machine-serial-number': 'SN-CPC2',
    })

    return faked_session


def test_option_help():
    """
    Test 'zhmc --help'
    """

    rc, stdout, stderr = call_zhmc_child(['--help'])

    assert_rc(0, rc, stdout, stderr)
    assert stdout.startswith(
        "Usage: zhmc [GENERAL-OPTIONS] COMMAND [ARGS]...\n"), \
        f"stdout={stdout!r}"
    assert stderr == ""


def test_option_version():
    """
    Test 'zhmc --version'
    """

    rc, stdout, stderr = call_zhmc_child(['--version'])

    assert_rc(0, rc, stdout, stderr)
    assert re.match(r'^zhmc, version [0-9]+\.[0-9]+\.[0-9]+', stdout)
    assert stderr == ""


OPTION_FORMAT_TABLE_PROPS_TESTCASES = [
    ('table',
     # Order of properties must match:
     '+-------------------+----------+\n'
     '| Field Name        | Value    |\n'
     '|-------------------+----------|\n'
     '| api-major-version | {v[amaj]:<8} |\n'
     '| api-minor-version | {v[amin]:<8} |\n'
     '| hmc-name          | {v[hnam]:8} |\n'
     '| hmc-version       | {v[hver]:8} |\n'
     '+-------------------+----------+\n'),
    ('plain',
     # Order of properties must match:
     'Field Name         Value\n'
     'api-major-version  {v[amaj]}\n'
     'api-minor-version  {v[amin]}\n'
     'hmc-name           {v[hnam]}\n'
     'hmc-version        {v[hver]}\n'),
    ('simple',
     # Order of properties must match:
     'Field Name         Value\n'
     '-----------------  --------\n'
     'api-major-version  {v[amaj]}\n'
     'api-minor-version  {v[amin]}\n'
     'hmc-name           {v[hnam]}\n'
     'hmc-version        {v[hver]}\n'),
    ('psql',
     # Order of properties must match:
     '+-------------------+----------+\n'
     '| Field Name        | Value    |\n'
     '|-------------------+----------|\n'
     '| api-major-version | {v[amaj]:<8} |\n'
     '| api-minor-version | {v[amin]:<8} |\n'
     '| hmc-name          | {v[hnam]:8} |\n'
     '| hmc-version       | {v[hver]:8} |\n'
     '+-------------------+----------+\n'),
    ('rst',
     # Order of properties must match:
     '=================  ========\n'
     'Field Name         Value\n'
     '=================  ========\n'
     'api-major-version  {v[amaj]}\n'
     'api-minor-version  {v[amin]}\n'
     'hmc-name           {v[hnam]}\n'
     'hmc-version        {v[hver]}\n'
     '=================  ========\n'),
    ('mediawiki',
     # Order of properties must match:
     '{{| class="wikitable" style="text-align: left;"\n'
     '|+ <!-- caption -->\n'
     '|-\n'
     '! Field Name        !! Value\n'
     '|-\n'
     '| api-major-version || {v[amaj]}\n'
     '|-\n'
     '| api-minor-version || {v[amin]}\n'
     '|-\n'
     '| hmc-name          || {v[hnam]}\n'
     '|-\n'
     '| hmc-version       || {v[hver]}\n'
     '|}}\n'),
    ('html',
     # Order of properties must match:
     '<table>\n'
     '<thead>\n'
     '<tr><th>Field Name       </th><th>Value   </th></tr>\n'
     '</thead>\n'
     '<tbody>\n'
     '<tr><td>api-major-version</td><td>{v[amaj]:<8}</td></tr>\n'
     '<tr><td>api-minor-version</td><td>{v[amin]:<8}</td></tr>\n'
     '<tr><td>hmc-name         </td><td>{v[hnam]:8}</td></tr>\n'
     '<tr><td>hmc-version      </td><td>{v[hver]:8}</td></tr>\n'
     '</tbody>\n'
     '</table>\n'),
    ('latex',
     # Order of properties must match:
     '\\begin{{tabular}}{{ll}}\n'
     '\\hline\n'
     ' Field Name        & Value    \\\\\n'
     '\\hline\n'
     ' api-major-version & {v[amaj]:<8} \\\\\n'
     ' api-minor-version & {v[amin]:<8} \\\\\n'
     ' hmc-name          & {v[hnam]:8} \\\\\n'
     ' hmc-version       & {v[hver]:8} \\\\\n'
     '\\hline\n'
     '\\end{{tabular}}\n'),
]


@pytest.mark.parametrize(
    "out_format, exp_stdout_template",
    OPTION_FORMAT_TABLE_PROPS_TESTCASES
)
@pytest.mark.parametrize(
    # Transpose only affects metrics output, but not info output.
    # Transpose is accepted and ignored for all table output formats.
    "transpose_opt", [
        None,
        '-x',
        '--transpose',
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('hmc-name', '2.14.0', '2.20'),
    ]
)
def test_option_format_table_props(
        hmc_name, hmc_version, api_version, out_opt, transpose_opt, out_format,
        exp_stdout_template):
    """
    Test global options (-o, --output-format) and (-x, --transpose), for all
    table formats, with the 'zhmc info' command which displays properties.
    """

    faked_session = FakedSession(
        'fake-host', hmc_name, hmc_version, api_version,
        userid='fake-user')

    api_version_parts = [int(vp) for vp in api_version.split('.')]
    exp_values = {
        'hnam': hmc_name,
        'hver': hmc_version,
        'amaj': api_version_parts[0],
        'amin': api_version_parts[1],
    }
    exp_stdout = exp_stdout_template.format(v=exp_values)

    args = [out_opt, out_format]
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.append('info')

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(0, rc, stdout, stderr)
    assert stdout == exp_stdout
    assert stderr == ""


# pylint: disable=line-too-long
OPTION_FORMAT_TABLE_RES_TESTCASES = [
    ('table',
     """+--------+-----------+---------------+--------------+----------------+-----------------+-------------------------+----------------------+
| name   | status    | dpm-enabled   | se-version   | machine-type   | machine-model   | machine-serial-number   | description          |
|--------+-----------+---------------+--------------+----------------+-----------------+-------------------------+----------------------|
| {c1[name]:<6} | {c1[status]:<9} | {c1[dpm-enabled]!r:<13} | {c1[se-version]:<12} | {c1[machine-type]:<14} | {c1[machine-model]:<15} | {c1[machine-serial-number]:<23} | {c1[description]:<20} |
| {c2[name]:<6} | {c2[status]:<9} | {c2[dpm-enabled]!r:<13} | {c2[se-version]:<12} | {c2[machine-type]:<14} | {c2[machine-model]:<15} | {c2[machine-serial-number]:<23} | {c2[description]:<20} |
+--------+-----------+---------------+--------------+----------------+-----------------+-------------------------+----------------------+
"""),  # noqa: E501
    ('plain',
     """name    status     dpm-enabled    se-version    machine-type    machine-model    machine-serial-number    description
{c1[name]:<6}  {c1[status]:<9}  {c1[dpm-enabled]!r:<13}  {c1[se-version]:<12}  {c1[machine-type]:<14}  {c1[machine-model]:<15}  {c1[machine-serial-number]:<23}  {c1[description]}
{c2[name]:<6}  {c2[status]:<9}  {c2[dpm-enabled]!r:<13}  {c2[se-version]:<12}  {c2[machine-type]:<14}  {c2[machine-model]:<15}  {c2[machine-serial-number]:<23}  {c2[description]}
"""),  # noqa: E501
    ('simple',
     """name    status     dpm-enabled    se-version    machine-type    machine-model    machine-serial-number    description
------  ---------  -------------  ------------  --------------  ---------------  -----------------------  --------------------
{c1[name]:<6}  {c1[status]:<9}  {c1[dpm-enabled]!r:<13}  {c1[se-version]:<12}  {c1[machine-type]:<14}  {c1[machine-model]:<15}  {c1[machine-serial-number]:<23}  {c1[description]}
{c2[name]:<6}  {c2[status]:<9}  {c2[dpm-enabled]!r:<13}  {c2[se-version]:<12}  {c2[machine-type]:<14}  {c2[machine-model]:<15}  {c2[machine-serial-number]:<23}  {c2[description]}
"""),  # noqa: E501
    ('psql',
     """+--------+-----------+---------------+--------------+----------------+-----------------+-------------------------+----------------------+
| name   | status    | dpm-enabled   | se-version   | machine-type   | machine-model   | machine-serial-number   | description          |
|--------+-----------+---------------+--------------+----------------+-----------------+-------------------------+----------------------|
| {c1[name]:<6} | {c1[status]:<9} | {c1[dpm-enabled]!r:<13} | {c1[se-version]:<12} | {c1[machine-type]:<14} | {c1[machine-model]:<15} | {c1[machine-serial-number]:<23} | {c1[description]:<20} |
| {c2[name]:<6} | {c2[status]:<9} | {c2[dpm-enabled]!r:<13} | {c2[se-version]:<12} | {c2[machine-type]:<14} | {c2[machine-model]:<15} | {c2[machine-serial-number]:<23} | {c2[description]:<20} |
+--------+-----------+---------------+--------------+----------------+-----------------+-------------------------+----------------------+
"""),  # noqa: E501
    ('rst',
     """======  =========  =============  ============  ==============  ===============  =======================  ====================
name    status     dpm-enabled    se-version    machine-type    machine-model    machine-serial-number    description
======  =========  =============  ============  ==============  ===============  =======================  ====================
{c1[name]:<6}  {c1[status]:<9}  {c1[dpm-enabled]!r:<13}  {c1[se-version]:<12}  {c1[machine-type]:<14}  {c1[machine-model]:<15}  {c1[machine-serial-number]:<23}  {c1[description]}
{c2[name]:<6}  {c2[status]:<9}  {c2[dpm-enabled]!r:<13}  {c2[se-version]:<12}  {c2[machine-type]:<14}  {c2[machine-model]:<15}  {c2[machine-serial-number]:<23}  {c2[description]}
======  =========  =============  ============  ==============  ===============  =======================  ====================
"""),  # noqa: E501
    ('mediawiki',
     """{{| class="wikitable" style="text-align: left;"
|+ <!-- caption -->
|-
! name   !! status    !! dpm-enabled   !! se-version   !! machine-type   !! machine-model   !! machine-serial-number   !! description
|-
| {c1[name]:<6} || {c1[status]:<9} || {c1[dpm-enabled]!r:<13} || {c1[se-version]:<12} || {c1[machine-type]:<14} || {c1[machine-model]:<15} || {c1[machine-serial-number]:<23} || {c1[description]}
|-
| {c2[name]:<6} || {c2[status]:<9} || {c2[dpm-enabled]!r:<13} || {c2[se-version]:<12} || {c2[machine-type]:<14} || {c2[machine-model]:<15} || {c2[machine-serial-number]:<23} || {c2[description]}
|}}
"""),  # noqa: E501
    ('html',
     """<table>
<thead>
<tr><th>name  </th><th>status   </th><th>dpm-enabled  </th><th>se-version  </th><th>machine-type  </th><th>machine-model  </th><th>machine-serial-number  </th><th>description         </th></tr>
</thead>
<tbody>
<tr><td>{c1[name]:<6}</td><td>{c1[status]:<9}</td><td>{c1[dpm-enabled]!r:<13}</td><td>{c1[se-version]:<12}</td><td>{c1[machine-type]:<14}</td><td>{c1[machine-model]:<15}</td><td>{c1[machine-serial-number]:<23}</td><td>{c1[description]:<20}</td></tr>
<tr><td>{c2[name]:<6}</td><td>{c2[status]:<9}</td><td>{c2[dpm-enabled]!r:<13}</td><td>{c2[se-version]:<12}</td><td>{c2[machine-type]:<14}</td><td>{c2[machine-model]:<15}</td><td>{c2[machine-serial-number]:<23}</td><td>{c2[description]:<20}</td></tr>
</tbody>
</table>
"""),  # noqa: E501
    ('latex',
     """\\begin{{tabular}}{{llllllll}}
\\hline
 name   & status    & dpm-enabled   & se-version   & machine-type   & machine-model   & machine-serial-number   & description          \\\\
\\hline
 {c1[name]:<6} & {c1[status]:<9} & {c1[dpm-enabled]!r:<13} & {c1[se-version]:<12} & {c1[machine-type]:<14} & {c1[machine-model]:<15} & {c1[machine-serial-number]:<23} & {c1[description]:<20} \\\\
 {c2[name]:<6} & {c2[status]:<9} & {c2[dpm-enabled]!r:<13} & {c2[se-version]:<12} & {c2[machine-type]:<14} & {c2[machine-model]:<15} & {c2[machine-serial-number]:<23} & {c2[description]:<20} \\\\
\\hline
\\end{{tabular}}
"""),  # noqa: E501
]
# pylint: enable=line-too-long


@pytest.mark.parametrize(
    "out_format, exp_stdout_template",
    OPTION_FORMAT_TABLE_RES_TESTCASES
)
@pytest.mark.parametrize(
    # Transpose only affects metrics output, but not info output.
    # Transpose is accepted and ignored for all table output formats.
    "transpose_opt", [
        None,
        '-x',
        '--transpose',
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
def test_option_format_table_res(
        out_opt, transpose_opt, out_format, exp_stdout_template):
    """
    Test global options (-o, --output-format) and (-x, --transpose), for all
    table formats, with the 'zhmc cpc list' command which displays a list of
    resources.
    """

    faked_session = setup_faked_hmc()
    client = Client(faked_session)
    cpcs = client.cpcs.list(full_properties=True)
    cpc1 = cpcs[0]
    cpc2 = cpcs[1]

    exp_stdout = exp_stdout_template.format(
        c1=cpc1.properties, c2=cpc2.properties)

    args = [out_opt, out_format]
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.extend(['cpc', 'list'])

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(0, rc, stdout, stderr)
    assert stdout == exp_stdout, (
        "Unexpected stdout:\n"
        f"Actual: {stdout!r}\n"
        f"Expected: {exp_stdout!r}")
    assert stderr == ""


# pylint: disable=line-too-long
OPTION_FORMAT_TABLE_DICT_TESTCASES = [
    ('table',
     """+-------------+-----------------------------------------+-------------+--------------------+-------------+------------+------------------+
| ec-number   | description                             | retrieved   | installable-conc   | activated   | accepted   | removable-conc   |
|-------------+-----------------------------------------+-------------+--------------------+-------------+------------+------------------|
| {f1[ec-number]:<11} | {f1[description]:<39} | {f1[retrieved]:<11} | {f1[installable-conc]:<18} | {f1[activated]:<11} | {f1[accepted]:<10} | {f1[removable-conc]:<16} |
| {f2[ec-number]:<11} | {f2[description]:<39} | {f2[retrieved]:<11} | {f2[installable-conc]:<18} | {f2[activated]:<11} | {f2[accepted]:<10} | {f2[removable-conc]:<16} |
+-------------+-----------------------------------------+-------------+--------------------+-------------+------------+------------------+
"""),  # noqa: E501
    ('plain',
     """ec-number    description                              retrieved    installable-conc    activated    accepted    removable-conc
{f1[ec-number]:<11}  {f1[description]:<39}  {f1[retrieved]:<11}  {f1[installable-conc]:<18}  {f1[activated]:<11}  {f1[accepted]:<10}  {f1[removable-conc]}
{f2[ec-number]:<11}  {f2[description]:<39}  {f2[retrieved]:<11}  {f2[installable-conc]:<18}  {f2[activated]:<11}  {f2[accepted]:<10}  {f2[removable-conc]}
"""),  # noqa: E501
    ('simple',
     """ec-number    description                              retrieved    installable-conc    activated    accepted    removable-conc
-----------  ---------------------------------------  -----------  ------------------  -----------  ----------  ----------------
{f1[ec-number]:<11}  {f1[description]:<39}  {f1[retrieved]:<11}  {f1[installable-conc]:<18}  {f1[activated]:<11}  {f1[accepted]:<10}  {f1[removable-conc]}
{f2[ec-number]:<11}  {f2[description]:<39}  {f2[retrieved]:<11}  {f2[installable-conc]:<18}  {f2[activated]:<11}  {f2[accepted]:<10}  {f2[removable-conc]}
"""),  # noqa: E501
    ('psql',
     """+-------------+-----------------------------------------+-------------+--------------------+-------------+------------+------------------+
| ec-number   | description                             | retrieved   | installable-conc   | activated   | accepted   | removable-conc   |
|-------------+-----------------------------------------+-------------+--------------------+-------------+------------+------------------|
| {f1[ec-number]:<11} | {f1[description]:<39} | {f1[retrieved]:<11} | {f1[installable-conc]:<18} | {f1[activated]:<11} | {f1[accepted]:<10} | {f1[removable-conc]:<16} |
| {f2[ec-number]:<11} | {f2[description]:<39} | {f2[retrieved]:<11} | {f2[installable-conc]:<18} | {f2[activated]:<11} | {f2[accepted]:<10} | {f2[removable-conc]:<16} |
+-------------+-----------------------------------------+-------------+--------------------+-------------+------------+------------------+
"""),  # noqa: E501
    ('rst',
     """===========  =======================================  ===========  ==================  ===========  ==========  ================
ec-number    description                              retrieved    installable-conc    activated    accepted    removable-conc
===========  =======================================  ===========  ==================  ===========  ==========  ================
{f1[ec-number]:<11}  {f1[description]:<39}  {f1[retrieved]:<11}  {f1[installable-conc]:<18}  {f1[activated]:<11}  {f1[accepted]:<10}  {f1[removable-conc]}
{f2[ec-number]:<11}  {f2[description]:<39}  {f2[retrieved]:<11}  {f2[installable-conc]:<18}  {f2[activated]:<11}  {f2[accepted]:<10}  {f2[removable-conc]}
===========  =======================================  ===========  ==================  ===========  ==========  ================
"""),  # noqa: E501
    ('mediawiki',
     """{{| class="wikitable" style="text-align: left;"
|+ <!-- caption -->
|-
! ec-number   !! description                             !! retrieved   !! installable-conc   !! activated   !! accepted   !! removable-conc
|-
| {f1[ec-number]:<11} || {f1[description]:<39} || {f1[retrieved]:<11} || {f1[installable-conc]:<18} || {f1[activated]:<11} || {f1[accepted]:<10} || {f1[removable-conc]}
|-
| {f2[ec-number]:<11} || {f2[description]:<39} || {f2[retrieved]:<11} || {f2[installable-conc]:<18} || {f2[activated]:<11} || {f2[accepted]:<10} || {f2[removable-conc]}
|}}
"""),  # noqa: E501
    ('html',
     """<table>
<thead>
<tr><th>ec-number  </th><th>description                            </th><th>retrieved  </th><th>installable-conc  </th><th>activated  </th><th>accepted  </th><th>removable-conc  </th></tr>
</thead>
<tbody>
<tr><td>{f1[ec-number]:<11}</td><td>{f1[description]:<39}</td><td>{f1[retrieved]:<11}</td><td>{f1[installable-conc]:<18}</td><td>{f1[activated]:<11}</td><td>{f1[accepted]:<10}</td><td>{f1[removable-conc]:<16}</td></tr>
<tr><td>{f2[ec-number]:<11}</td><td>{f2[description]:<39}</td><td>{f2[retrieved]:<11}</td><td>{f2[installable-conc]:<18}</td><td>{f2[activated]:<11}</td><td>{f2[accepted]:<10}</td><td>{f2[removable-conc]:<16}</td></tr>
</tbody>
</table>
"""),  # noqa: E501
    ('latex',
     """\\begin{{tabular}}{{lllllll}}
\\hline
 ec-number   & description                             & retrieved   & installable-conc   & activated   & accepted   & removable-conc   \\\\
\\hline
 {f1[ec-number]:<11} & {f1[description]:<39} & {f1[retrieved]:<11} & {f1[installable-conc]:<18} & {f1[activated]:<11} & {f1[accepted]:<10} & {f1[removable-conc]:<16} \\\\
 {f2[ec-number]:<11} & {f2[description]:<39} & {f2[retrieved]:<11} & {f2[installable-conc]:<18} & {f2[activated]:<11} & {f2[accepted]:<10} & {f2[removable-conc]:<16} \\\\
\\hline
\\end{{tabular}}
"""),  # noqa: E501
]
# pylint: enable=line-too-long


@pytest.mark.parametrize(
    "out_format, exp_stdout_template",
    OPTION_FORMAT_TABLE_DICT_TESTCASES
)
@pytest.mark.parametrize(
    # Transpose only affects metrics output, but not info output.
    # Transpose is accepted and ignored for all table output formats.
    "transpose_opt", [
        None,
        '-x',
        '--transpose',
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
def test_option_format_table_dict(
        out_opt, transpose_opt, out_format, exp_stdout_template):
    """
    Test global options (-o, --output-format) and (-x, --transpose), for all
    table formats, with the 'zhmc console list-firmware' command which displays
    a dict.
    """

    faked_session = setup_faked_hmc()
    client = Client(faked_session)
    console = client.consoles.console

    console.pull_properties('ec-mcl-description')
    ec_mcl = console.properties['ec-mcl-description']
    firmware_list = convert_ec_mcl_description(ec_mcl)

    # The dicts display sorts the output by the first column
    sorted_firmware_list = sorted(
        firmware_list, key=lambda row: row[list(row.keys())[0]])
    f1 = sorted_firmware_list[0]
    f2 = sorted_firmware_list[1]

    exp_stdout = exp_stdout_template.format(f1=f1, f2=f2)

    args = [out_opt, out_format]
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.extend(['console', 'list-firmware'])

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(0, rc, stdout, stderr)
    assert stdout == exp_stdout, (
        "Unexpected stdout:\n"
        f"Actual: {stdout!r}\n"
        f"Expected: {exp_stdout!r}")
    assert stderr == ""


JSON_PROPS_STDOUT_TEMPLATE = \
    '{{' \
    '"api-major-version": {v[amaj]},' \
    '"api-minor-version": {v[amin]},' \
    '"hmc-name": "{v[hnam]}",' \
    '"hmc-version": "{v[hver]}"' \
    '}}'

JSON_PROPS_CONFLICT_PATTERNS = [
    r"Error: Transposing output tables .* conflicts with non-table "
    r"output format .* json",
]


@pytest.mark.parametrize(
    "transpose_opt, exp_rc, exp_stdout_template, exp_stderr_patterns", [
        (None, 0, JSON_PROPS_STDOUT_TEMPLATE, None),
        ('-x', 1, None, JSON_PROPS_CONFLICT_PATTERNS),
        ('--transpose', 1, None, JSON_PROPS_CONFLICT_PATTERNS),
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('hmc-name', '2.14.0', '2.20'),
    ]
)
def test_option_format_json_props(
        hmc_name, hmc_version, api_version, out_opt, transpose_opt,
        exp_rc, exp_stdout_template, exp_stderr_patterns):
    """
    Test global options (-o, --output-format) and (-x, --transpose), for the
    'json' format, with the 'zhmc info' command which displays properties.
    """

    faked_session = FakedSession(
        'fake-host', hmc_name, hmc_version, api_version,
        userid='fake-user')

    args = [out_opt, 'json']
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.append('info')

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(exp_rc, rc, stdout, stderr)

    if exp_stdout_template:
        api_version_parts = [int(vp) for vp in api_version.split('.')]
        exp_values = {
            'hnam': hmc_name,
            'hver': hmc_version,
            'amaj': api_version_parts[0],
            'amin': api_version_parts[1],
        }
        exp_stdout = exp_stdout_template.format(v=exp_values)
        exp_stdout_dict = json.loads(exp_stdout)
        stdout_dict = json.loads(stdout)
        assert stdout_dict == exp_stdout_dict
    else:
        assert stdout == ""

    if exp_stderr_patterns:
        assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')
    else:
        assert stderr == ""


JSON_RES_STDOUT_TEMPLATE = \
    '[' \
    '{{' \
    '"name": "{c1[name]}",' \
    '"status": "{c1[status]}",' \
    '"dpm-enabled": {c1[dpm-enabled]},' \
    '"se-version": "{c1[se-version]}",' \
    '"machine-type": "{c1[machine-type]}",' \
    '"machine-model": "{c1[machine-model]}",' \
    '"machine-serial-number": "{c1[machine-serial-number]}",' \
    '"description": "{c1[description]}"' \
    '}},' \
    '{{' \
    '"name": "{c2[name]}",' \
    '"status": "{c2[status]}",' \
    '"dpm-enabled": {c2[dpm-enabled]},' \
    '"se-version": "{c2[se-version]}",' \
    '"machine-type": "{c2[machine-type]}",' \
    '"machine-model": "{c2[machine-model]}",' \
    '"machine-serial-number": "{c2[machine-serial-number]}",' \
    '"description": "{c2[description]}"' \
    '}}' \
    ']'

JSON_RES_CONFLICT_PATTERNS = [
    r"Error: Transposing output tables .* conflicts with non-table "
    r"output format .* json",
]


@pytest.mark.parametrize(
    "transpose_opt, exp_rc, exp_stdout_template, exp_stderr_patterns", [
        (None, 0, JSON_RES_STDOUT_TEMPLATE, None),
        ('-x', 1, None, JSON_RES_CONFLICT_PATTERNS),
        ('--transpose', 1, None, JSON_RES_CONFLICT_PATTERNS),
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
def test_option_format_json_res(
        out_opt, transpose_opt, exp_rc, exp_stdout_template,
        exp_stderr_patterns):
    """
    Test global options (-o, --output-format) and (-x, --transpose), for the
    'json' format, with the 'zhmc cpc list' command which displays a list of
    resources.
    """

    faked_session = setup_faked_hmc()
    client = Client(faked_session)
    cpcs = client.cpcs.list(full_properties=True)
    cpc1 = cpcs[0]
    cpc2 = cpcs[1]

    args = [out_opt, 'json']
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.extend(['cpc', 'list'])

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(exp_rc, rc, stdout, stderr)

    if exp_stdout_template:
        c1 = dict(cpc1.properties)
        c1['dpm-enabled'] = 'true' if c1['dpm-enabled'] else 'false'
        c2 = dict(cpc2.properties)
        c2['dpm-enabled'] = 'true' if c2['dpm-enabled'] else 'false'
        exp_stdout = exp_stdout_template.format(c1=c1, c2=c2)
        exp_stdout_dict = json.loads(exp_stdout)
        stdout_dict = json.loads(stdout)
        assert stdout_dict == exp_stdout_dict, (
            "Unexpected stdout (as dict):\n"
            f"Actual: {stdout_dict!r}\n"
            f"Expected: {exp_stdout_dict!r}")
    else:
        assert stdout == ""

    if exp_stderr_patterns:
        assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')
    else:
        assert stderr == ""


JSON_DICT_STDOUT_TEMPLATE = \
    '[' \
    '{{' \
    '"ec-number": "{f1[ec-number]}",' \
    '"description": "{f1[description]}",' \
    '"retrieved": "{f1[retrieved]}",' \
    '"installable-conc": "{f1[installable-conc]}",' \
    '"activated": "{f1[activated]}",' \
    '"accepted": "{f1[accepted]}",' \
    '"removable-conc": "{f1[removable-conc]}"' \
    '}},' \
    '{{' \
    '"ec-number": "{f2[ec-number]}",' \
    '"description": "{f2[description]}",' \
    '"retrieved": "{f2[retrieved]}",' \
    '"installable-conc": "{f2[installable-conc]}",' \
    '"activated": "{f2[activated]}",' \
    '"accepted": "{f2[accepted]}",' \
    '"removable-conc": "{f2[removable-conc]}"' \
    '}}' \
    ']'

JSON_DICT_CONFLICT_PATTERNS = [
    r"Error: Transposing output tables .* conflicts with non-table "
    r"output format .* json",
]


@pytest.mark.parametrize(
    "transpose_opt, exp_rc, exp_stdout_template, exp_stderr_patterns", [
        (None, 0, JSON_DICT_STDOUT_TEMPLATE, None),
        ('-x', 1, None, JSON_DICT_CONFLICT_PATTERNS),
        ('--transpose', 1, None, JSON_DICT_CONFLICT_PATTERNS),
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
def test_option_format_json_dict(
        out_opt, transpose_opt, exp_rc, exp_stdout_template,
        exp_stderr_patterns):
    """
    Test global options (-o, --output-format) and (-x, --transpose), for the
    'json' format, with the 'zhmc cpc list' command which displays a list of
    resources.
    """

    faked_session = setup_faked_hmc()
    client = Client(faked_session)
    console = client.consoles.console

    console.pull_properties('ec-mcl-description')
    ec_mcl = console.properties['ec-mcl-description']
    firmware_list = convert_ec_mcl_description(ec_mcl)

    # The JSON dicts display does not sort the output
    f1 = firmware_list[0]
    f2 = firmware_list[1]

    args = [out_opt, 'json']
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.extend(['console', 'list-firmware'])

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(exp_rc, rc, stdout, stderr)

    if exp_stdout_template:
        exp_stdout = exp_stdout_template.format(f1=f1, f2=f2)
        exp_stdout_dict = json.loads(exp_stdout)
        stdout_dict = json.loads(stdout)
        assert stdout_dict == exp_stdout_dict, (
            "Unexpected stdout (as dict):\n"
            f"Actual: {stdout_dict!r}\n"
            f"Expected: {exp_stdout_dict!r}")
    else:
        assert stdout == ""

    if exp_stderr_patterns:
        assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')
    else:
        assert stderr == ""


CSV_RES_STDOUT_TEMPLATE = (
    # Note: The quoting must be consistent with how the CSV file is written.
    '"name","status","dpm-enabled","se-version","machine-type",'
    '"machine-model","machine-serial-number","description"\n'
    '"{c1[name]}","{c1[status]}",{c1[dpm-enabled]},"{c1[se-version]}",'
    '"{c1[machine-type]}","{c1[machine-model]}","{c1[machine-serial-number]}",'
    '"{c1[description]}"\n'
    '"{c2[name]}","{c2[status]}",{c2[dpm-enabled]},"{c2[se-version]}",'
    '"{c2[machine-type]}","{c2[machine-model]}","{c2[machine-serial-number]}",'
    '"{c2[description]}"\n'
)

CSV_RES_CONFLICT_PATTERNS = [
    r"Error: Transposing output tables .* conflicts with non-table "
    r"output format .* csv",
]


@pytest.mark.parametrize(
    "transpose_opt, exp_rc, exp_stdout_template, exp_stderr_patterns", [
        (None, 0, CSV_RES_STDOUT_TEMPLATE, None),
        ('-x', 1, None, CSV_RES_CONFLICT_PATTERNS),
        ('--transpose', 1, None, CSV_RES_CONFLICT_PATTERNS),
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
def test_option_format_csv_res(
        out_opt, transpose_opt, exp_rc, exp_stdout_template,
        exp_stderr_patterns):
    """
    Test global options (-o, --output-format) and (-x, --transpose),
    for the 'csv' output format,
    with the 'zhmc cpc list' command which displays a list of resources.
    """

    faked_session = setup_faked_hmc()
    client = Client(faked_session)
    cpcs = client.cpcs.list(full_properties=True)
    cpc1 = cpcs[0]
    cpc2 = cpcs[1]

    args = [out_opt, 'csv']
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.extend(['cpc', 'list'])

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(exp_rc, rc, stdout, stderr)

    if exp_stdout_template:
        exp_stdout = exp_stdout_template.format(
            c1=cpc1.properties, c2=cpc2.properties, q='"')
        assert stdout == exp_stdout, (
            "Unexpected stdout:\n"
            f"Actual: {stdout!r}\n"
            f"Expected: {exp_stdout!r}")
    else:
        assert stdout == ""

    if exp_stderr_patterns:
        assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')
    else:
        assert stderr == ""


CSV_DICT_STDOUT_TEMPLATE = (
    '"ec-number","description","retrieved","installable-conc","activated",'
    '"accepted","removable-conc"\n'
    '"{f1[ec-number]}","{f1[description]}","{f1[retrieved]}",'
    '"{f1[installable-conc]}","{f1[activated]}","{f1[accepted]}",'
    '"{f1[removable-conc]}"\n'
    '"{f2[ec-number]}","{f2[description]}","{f2[retrieved]}",'
    '"{f2[installable-conc]}","{f2[activated]}","{f2[accepted]}",'
    '"{f2[removable-conc]}"\n'
)


CSV_DICT_CONFLICT_PATTERNS = [
    r"Error: Transposing output tables .* conflicts with non-table "
    r"output format .* csv",
]


@pytest.mark.parametrize(
    "transpose_opt, exp_rc, exp_stdout_template, exp_stderr_patterns", [
        (None, 0, CSV_DICT_STDOUT_TEMPLATE, None),
        ('-x', 1, None, CSV_DICT_CONFLICT_PATTERNS),
        ('--transpose', 1, None, CSV_DICT_CONFLICT_PATTERNS),
    ]
)
@pytest.mark.parametrize(
    "out_opt", ['-o', '--output-format']
)
def test_option_format_csv_dict(
        out_opt, transpose_opt, exp_rc, exp_stdout_template,
        exp_stderr_patterns):
    """
    Test global options (-o, --output-format) and (-x, --transpose),
    for the 'csv' output format,
    with the 'zhmc cpc list' command which displays a list of resources.
    """

    faked_session = setup_faked_hmc()
    client = Client(faked_session)
    console = client.consoles.console

    console.pull_properties('ec-mcl-description')
    ec_mcl = console.properties['ec-mcl-description']
    firmware_list = convert_ec_mcl_description(ec_mcl)

    # The dicts display sorts the output by the first column
    sorted_firmware_list = sorted(
        firmware_list, key=lambda row: row[list(row.keys())[0]])
    f1 = sorted_firmware_list[0]
    f2 = sorted_firmware_list[1]

    args = [out_opt, 'csv']
    if transpose_opt is not None:
        args.append(transpose_opt)
    args.extend(['console', 'list-firmware'])

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        args, faked_session=faked_session)

    assert_rc(exp_rc, rc, stdout, stderr)

    if exp_stdout_template:
        exp_stdout = exp_stdout_template.format(f1=f1, f2=f2, q='"')
        assert stdout == exp_stdout, (
            "Unexpected stdout:\n"
            f"Actual: {stdout!r}\n"
            f"Expected: {exp_stdout!r}")
    else:
        assert stdout == ""

    if exp_stderr_patterns:
        assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')
    else:
        assert stderr == ""


LOG_API_DEBUG_PATTERNS = [
    r".*DEBUG zhmcclient.api: .* Client.query_api_version\(\), "
    r"args: \(.*\), kwargs: \{.*\}",
    r".*DEBUG zhmcclient.api: .* Client.query_api_version\(\), "
    r"result: \{.*\}",
]


@pytest.mark.parametrize(
    "log_value, exp_rc, exp_stderr_patterns", [
        ('api=error', 0, []),
        ('api=warning', 0, []),
        ('api=info', 0, []),
        ('api=debug', 0, LOG_API_DEBUG_PATTERNS),
        ('api=debug,hmc=error', 0, LOG_API_DEBUG_PATTERNS),
        (',api=debug,hmc=error', 0, LOG_API_DEBUG_PATTERNS),
        ('api=debug,,hmc=error', 0, LOG_API_DEBUG_PATTERNS),
        ('api=debug,hmc=error,', 0, LOG_API_DEBUG_PATTERNS),
        (',,api=debug,,', 0, LOG_API_DEBUG_PATTERNS),
        (',,', 0, []),
        ('api:debug', 1, ["Error: Missing '=' .*"]),
        ('api=debugx', 1, ["Error: Invalid log level .*"]),
        ('apix=debug', 1, ["Error: Invalid log component .*"]),
    ]
)
@pytest.mark.parametrize(
    "log_opt", ['--log']
)
@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('hmc-name', '2.14.0', '10.2'),
    ]
)
def test_option_log(
        hmc_name, hmc_version, api_version, log_opt, log_value,
        exp_rc, exp_stderr_patterns):
    """
    Test 'zhmc info' with global option --log
    """

    faked_session = FakedSession(
        'fake-host', hmc_name, hmc_version, api_version,
        userid='fake-user')

    # Invoke the command to be tested
    rc, stdout, stderr = call_zhmc_inline(
        [log_opt, log_value, 'info'],
        faked_session=faked_session)

    assert_rc(exp_rc, rc, stdout, stderr)
    assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')


@pytest.mark.parametrize(
    "logdest_value, exp_rc, exp_stderr_patterns", [
        (None, 0, LOG_API_DEBUG_PATTERNS),
        ('stderr', 0, LOG_API_DEBUG_PATTERNS),
        ('syslog', 0, []),
        ('none', 0, []),
        (TEST_LOGFILE, 0, []),
    ]
)
@pytest.mark.parametrize(
    "logdest_opt", ['--log-dest']
)
@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('fake-hmc', '2.14.0', '10.2'),
    ]
)
def test_option_logdest(
        hmc_name, hmc_version, api_version, logdest_opt,
        logdest_value, exp_rc, exp_stderr_patterns):
    """
    Test 'zhmc info' with global option --log-dest (and --log)
    """

    faked_session = FakedSession(
        'fake-host', hmc_name, hmc_version, api_version,
        userid='fake-user')

    args = ['--log', 'api=debug']
    logger_name = 'zhmcclient.api'  # corresponds to --log option
    if logdest_value is not None:
        args.append(logdest_opt)
        args.append(logdest_value)
    args.append('info')

    try:

        # Remove a possibly existing log file
        if logdest_value == TEST_LOGFILE:
            if os.path.exists(TEST_LOGFILE):
                os.remove(TEST_LOGFILE)

        # Invoke the command to be tested
        rc, stdout, stderr = call_zhmc_inline(
            args, faked_session=faked_session)

        assert_rc(exp_rc, rc, stdout, stderr)
        assert_patterns(exp_stderr_patterns, stderr.splitlines(), 'stderr')

        # Check system log
        if logdest_value == 'syslog':
            syslog_files = ['/var/log/messages', '/var/log/syslog']
            for syslog_file in syslog_files:
                if os.path.exists(syslog_file):
                    break
            else:
                syslog_file = None
                print("Warning: Cannot check syslog; syslog file not found "
                      "in: {f!r}".format(f=syslog_files))
            syslog_lines = None
            if syslog_file:
                try:
                    syslog_lines = subprocess.check_output(
                        'sudo tail {f} || tail {f}'.format(f=syslog_file),
                        shell=True)  # nosec: B602
                except Exception as exc:  # pylint: disable=broad-except
                    print("Warning: Cannot tail syslog file {f}: {msg}".
                          format(f=syslog_file, msg=exc))
            if syslog_lines:
                syslog_lines = syslog_lines.decode('utf-8').splitlines()
                logger_lines = []
                for line in syslog_lines:
                    if logger_name in line:
                        logger_lines.append(line)
                logger_lines = logger_lines[
                    -len(LOG_API_DEBUG_PATTERNS):]
                exp_patterns = [r'.*' + p for p in LOG_API_DEBUG_PATTERNS]
                assert_patterns(exp_patterns, logger_lines, 'syslog')

        # Check log file
        if logdest_value == TEST_LOGFILE:
            with open(TEST_LOGFILE, encoding='utf-8') as fp:
                log_lines = fp.readlines()
                logger_lines = []
                for line in log_lines:
                    if logger_name in line:
                        logger_lines.append(line)
                logger_lines = logger_lines[
                    -len(LOG_API_DEBUG_PATTERNS):]
                exp_patterns = [r'.*' + p for p in LOG_API_DEBUG_PATTERNS]
                assert_patterns(exp_patterns, logger_lines, 'syslog')

    finally:
        # Clean up a possibly existing log file
        if logdest_value == TEST_LOGFILE:
            if os.path.exists(TEST_LOGFILE):
                try:
                    os.remove(TEST_LOGFILE)
                except OSError:
                    # On Windows with Python 3, PermissionError is raised.
                    # TODO: Find out why and resolve this better.
                    pass
